/* $Id: caf_find_misassemblies.c,v 1.3 2005/06/16 11:42:37 rmd Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "consensus.h"
#include "cl.h"
#include "caf_find_misassemblies.h"

int iterate_contig(cafAssembly* caf, CAFSEQ* seq, Array tag_info,
                   do_iterator_callback callback,
                   void *callback_params, FILE *log);

int main (int argc, char **argv)
{
  FILE *log=NULL;
  cafAssembly *CAF;
  char CAFfile[256] = {0};
  char logfile[100] = {0};
  float bad = BAD;
  int high = HIGH;
  int help = 0;

  getint("-high", &high, argc, argv);
  getfloat("-bad", &bad, argc, argv);
  getboolean("-h", &help, argc, argv);
  log = argfile("-log","w",argc,argv,logfile);

  if(help) 
    usage();
  

  if ((CAF = readCAF(argc,argv,CAFfile)) != 0 )
      find_misassemblies(CAF,bad,high,log);
  
  writeCAF( CAF, CAFfile, argc, argv);
  
  if(log)
    fclose(log);
  
  return 0;
}


int find_misassemblies (cafAssembly * caf, double bad, int high, FILE *log)
{
  Array   seqs;
  Array   tag_info;
  CAFSEQ *seq;
  int i;

  seqs = caf->seqs;
  
  tag_info = arrayCreate(arrayMax(caf->seqs), To_tag);
  arrayp(tag_info, arrayMax(caf->seqs), To_tag);

  for (i = 0; i < arrayMax(seqs); i++) {
    seq = arrp(seqs, i, CAFSEQ);      
    if (seq->type == is_contig) {
      fprintf(stderr,"%s ",dictName(caf->seqDict,seq->id));
      iterate_contig(caf, seq, tag_info, check_base, 0, log);
    }
  }

  tag_misassemblies(caf, tag_info, arrayMax(caf->seqs), log);
  
  arrayDestroy(tag_info);

  return 0;
}

int iterate_contig(cafAssembly* caf, CAFSEQ* seq, Array tag_info,
                   do_iterator_callback callback,
                   void *callback_params, FILE *log)
{
  ASSINF *assembled_froms; /* First ASSINF struct */
  ASSINF *last_af;         /* Last  ASSINF struct */
  ASSINF *curr_af;         /* pointer to current ASSINF struct        */
  ASSINF *temp_af;
  To_tag *tPtr;
  Array  segments;         /* Array holding all Seg_entry at current base */ 
  int    start=0;          /* First good base in contig consensus     */
  int    end=0;            /* Last  good base in contig consensus + 1 */
  int    i;                /* The current consensus base */
  int    k;
  
  /* Check if this contig has any assembled_from information */
  if (!seq->assinf)               return 0;
  if (arrayMax(seq->assinf) == 0) return 0;
  
  /* Sort the assembled_from information into order along the contig */
  
  assembled_froms = arrp(seq->assinf, 0, ASSINF);
  qsort(assembled_froms, arrayMax(seq->assinf), sizeof(ASSINF), af_compare);
  
  /* Find the first and last bases in the assembled_froms */
  for(k=0;k<arrayMax(seq->assinf);k++) {
    temp_af = arrp(seq->assinf,k,ASSINF);
    /* Initialise struct counters to 0 and assign name. */
    tPtr = arrp(tag_info,temp_af->seq,To_tag);    
    tPtr->align = 0;
    tPtr->match = 0;
    tPtr->read = temp_af->seq;
    if(temp_af->s1 > temp_af->s2) {
      if(start > temp_af->s2) { start = temp_af->s2; }
      if(end < temp_af->s1) { end = temp_af->s1; }
    }else{
      if(start > temp_af->s1) { start = temp_af->s1; }
      if(end < temp_af->s2) { end = temp_af->s2; }
    }
  }

  last_af = arrp(seq->assinf, arrayMax(seq->assinf)-1, ASSINF);

  if (start < 0) return 1;
  
  /* Go through each consensus base, work out which read segments are
     contributing to the base */

  fprintf(stderr,"%d..%d\n",start,end);
  
  curr_af  = assembled_froms;
  segments = arrayCreate(128, Seg_entry);
  
  for (i = start; i < end; i++) {
    update_segments(caf, i, segments, &curr_af, last_af);    
    (*callback)(caf, seq, i, segments, tag_info, callback_params, log);    
  }  
  
  arrayDestroy(segments);
               
  return 0;
}
/* Tag misassembled reads */
static int tag_misassemblies(cafAssembly *caf, Array tag_info, int rnum, FILE * log) 
{
  CAFSEQ *rPtr;
  To_tag *tpr;
  int i;
  double percent;
  char string[100] = {0};
  
  for(i=0;i<arrayMax(caf->seqs);i++) {
    rPtr = arrp(caf->seqs,i,CAFSEQ);
    if(rPtr->type == is_read) {
      tpr = arrp(tag_info,i,To_tag);
      if(tpr->align==0)
        tpr->align = 1;
      
      percent =  ((float)tpr->match / (float)tpr->align) * 100;
      
      if(percent < BAD ) {
        sprintf(string,"Percent match %.1f",percent);
        newTag(rPtr,caf,"MISS",1,rPtr->len,string);
        
        if(log)
          fprintf(log,"%s misaligned %.1f%% match\n", dictName(caf->seqDict,rPtr->id), percent);              
      } 
    }
  }
  return 0;
}


static void update_segments(cafAssembly *caf, int i, Array segments,
                            ASSINF **curr_af, ASSINF *last_af)
{
  Seg_entry *new_seg; /* New segment */
  int seg;       /* Segment counter */
  int inserted;  /* Flag if a new segment has been inserted */
  
  /* Remove any segments we have finished with (i.e. i has gone past the end).
     As the segments array is sorted by segment end, the ones we need to get
     rid of should be at the end of the array */

  for (seg = arrayMax(segments) - 1; seg >= 0; seg--) {
    
    if (arrp(segments, seg, Seg_entry)->ce > i) break;

    arrayMax(segments)--;
  }
  
  inserted = 0;

  /* Go through the remaining ASSINF entries and see if they should be put
     into the segments array yet. */

  for (; (*curr_af) <= last_af; (*curr_af)++) {

    if (min((*curr_af)->s1, (*curr_af)->s2) - 1 > i) break;

    /* Insert the next segment */

    new_seg = arrayp(segments, arrayMax(segments), Seg_entry);

    if ((*curr_af)->s1 < (*curr_af)->s2) {  /* Forward strand */

      new_seg->read = arrp(caf->seqs, (*curr_af)->seq, CAFSEQ);
      new_seg->cs   = (*curr_af)->s1 - 1;
      new_seg->ce   = (*curr_af)->s2;
      new_seg->rs   = (*curr_af)->r1 - 1;
      new_seg->re   = (*curr_af)->r2;
      new_seg->comp = 0;
    } else {                          /* Reverse strand */
      
      new_seg->read = arrp(caf->seqs, (*curr_af)->seq, CAFSEQ);
      new_seg->cs   = (*curr_af)->s2 - 1;
      new_seg->ce   = (*curr_af)->s1;
      new_seg->rs   = (*curr_af)->r1 - 1;
      new_seg->re   = (*curr_af)->r2;
      new_seg->comp = 1;
    }
    inserted = 1;
  }

  if (inserted) {

    /* re-sort the array of segments by end position */

    qsort(arrp(segments, 0, Seg_entry), arrayMax(segments),
          sizeof(Seg_entry), seg_compare);
  }
}


static void check_base(cafAssembly* caf, CAFSEQ* contig,
		       int i, Array segments, Array tag_info, void * params,
		       FILE *log)
{
  
  int seg, rp, j = 0, k = 0, hit = 0, q = 0;
  char base,qual;
  char BASE[256]     = {0};
  char string[100]   = {0};
  Array h_quality;
  Seg_entry *se;
  high_qual *new_qual;  
  To_tag *tag;

  h_quality = arrayCreate(128, high_qual);

  for (seg = 0; seg < arrayMax(segments); seg++) {
    se = arrp(segments, seg, Seg_entry);

    tag = arrp(tag_info,se->read->id,To_tag);
    tag->align++;
    
    if (se->comp) { /* reverse strand */   
      rp = se->re - (i - se->cs) - 1;
      base = (char) toupper((int) complement(*(stackText(se->read->dna, rp))));
      qual = (int) arr(se->read->base_quality, rp, char);      
    } else {        /* forward strand */      
      rp = i - se->cs + se->rs;
      base = (char) toupper((int) *(stackText(se->read->dna, rp)));
      qual = (int) arr(se->read->base_quality, rp, char);      
    }    
    if(base == *stackText(contig->dna, i)) {
      tag->match++;
    }else{
      if(qual >= HIGH) {         
        new_qual = arrayp(h_quality, arrayMax(h_quality), high_qual);
        new_qual->rs = se->read;
        new_qual->qual = qual;  
        new_qual->pos = rp+1;
      }
    }    
    if(BASE[toupper((int) base)] == 0 || BASE[toupper((int) base)] < qual) {
      BASE[toupper((int) base)] = qual;
    }
  }
  
  for(j=0;j<256;j++) {
    if(BASE[j] > 0) {
      k++;      
    }
  }
  if(k > 1) {
    for(j=0;j<256;j++) {
      if(BASE[j] >= HIGH) {
        hit++;      
      }
    }
  }
  
  if(hit > 1){
    for (q = 0; q < arrayMax(h_quality); q++) {
      new_qual = arrp(h_quality,q,high_qual);     
      sprintf(string,"%d",new_qual->qual);
      newTag(new_qual->rs,caf,"AMBG",new_qual->pos,new_qual->pos,string);
    }
  }
  
  arrayDestroy(h_quality);
}

void usage (void)
{
  fprintf(stderr,"\nUsage: caf_find_misassemblies [-h] [options] < caf.in > caf.out\n\n");
  exit(0);
}

static int af_compare(const void *a, const void *b)
{
  int mina,minb;
  ASSINF *aa = (ASSINF *) a;
  ASSINF *bb = (ASSINF *) b;
  
  mina = min(aa->s1, aa->s2);
  minb = min(bb->s1, bb->s2);


  return (mina - minb);
}

/* Compare Seg_entry entries.  When used for sorting, they will be ordered by
   end position (ce), high to low */

static int seg_compare(const void *a, const void *b)
{
  Seg_entry *aa = (Seg_entry *) a;
  Seg_entry *bb = (Seg_entry *) b;

  return (bb->ce - aa->ce);
}
static char complement(char b)
{
  static char tab[256] = {0};
  int i;

  if (!tab['a']) {
    for (i = 0; i < 256; i++) tab[i] = 'N';
    tab['a'] = 't';
    tab['A'] = 'T';
    tab['c'] = 'g';
    tab['C'] = 'G';
    tab['g'] = 'c';
    tab['G'] = 'C';
    tab['t'] = 'a';
    tab['T'] = 'A';
    tab['n'] = 'n';
    tab['-'] = '-';
  }

  return tab[(int) b];
}
